#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/sctp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/types.h>
#include <netdb.h>
#include <pthread.h>
#include <errno.h>

#define ROW 0
#define NUM 0

#define BUFFERSIZE 800
#define LOOPS 1

static int create_assocs()
{
	int fd, i, j;
	char addr_str[30];
	struct addrinfo *res = NULL;
	struct sctp_event_subscribe evnts;
	struct sockaddr_storage addr_cli, addr_ser;

	memset(&addr_ser, 0, sizeof(addr_ser));
	sprintf(addr_str, "172.16.%d.%d", 254, 254);
	getaddrinfo(addr_str, "8888", NULL, &res);
	memcpy(&addr_ser, res->ai_addr, res->ai_addrlen);
	((struct sockaddr_in *)(&addr_ser))->sin_port = htons(8888);

	for(i=1; i<=ROW; i++){
		for(j=1; j<=NUM; j++) {
			memset(&addr_cli, 0, sizeof(addr_cli));
			sprintf(addr_str, "172.16.%d.%d", i, j);
			getaddrinfo(addr_str, "8888", NULL, &res);
			memcpy(&addr_cli, res->ai_addr, res->ai_addrlen);
			((struct sockaddr_in *)(&addr_cli))->sin_port = htons(8888);

			if ((fd = socket(res->ai_family, SOCK_SEQPACKET, IPPROTO_SCTP)) == -1){
				perror("socket");
				exit(-1);
			}
			if (bind(fd, (struct sockaddr *)&addr_cli, sizeof(addr_cli)) == -1){
				perror("bind");
				exit(-1);
			}
			memset(&evnts, 0, sizeof(evnts));
			evnts.sctp_data_io_event = 1;
			if (setsockopt(fd, IPPROTO_SCTP, SCTP_EVENTS, &evnts,sizeof(evnts)) == -1){
				perror("setsockopt");
				exit(-1);
			}
			if (connect(fd, (struct sockaddr *)&addr_ser, sizeof(addr_ser)) == -1){
				perror("connect");
				exit(-1);
			}
		}
	}
	return 0;
}
static int get_asoc_id(int fd)
{
	char *buf;
	int len, asoc_id = 0;

	len = (1 * sizeof(sctp_assoc_t)) + sizeof(uint32_t);
	buf = (char *)malloc(len); 

	if (getsockopt(fd, IPPROTO_SCTP, SCTP_GET_ASSOC_ID_LIST, (void *)buf, &len) < 0) {
		perror("setsockopt");
		exit(-1);
	}

	asoc_id = ((struct sctp_assoc_ids *)buf)->gaids_assoc_id[0];
	return asoc_id;
}

int my_get_prsctp_support(int fd, int asoc_id)
{
	struct sctp_assoc_value params;
	socklen_t len;
	int result;

	len = sizeof(params);
	params.assoc_id = asoc_id;
	result = getsockopt(fd, IPPROTO_SCTP, SCTP_PR_SUPPORTED, &params, &len);
	if (result < 0) {
		perror("getsockopt");
		exit(-1);
	}
	if (asoc_id)
		printf("    Get ASOC %d\n", params.assoc_value);
	else
		printf("    Get EP %d\n", params.assoc_value);
}
int my_set_prsctp_support(int fd, int asoc_id, int asoc_value)
{
	struct sctp_assoc_value params;
	socklen_t len;
	int result;

	len = sizeof(params);
	params.assoc_id = asoc_id;
	params.assoc_value = asoc_value;
	printf("[Set Asoc ID %d, VA %d ]\n",params.assoc_id, params.assoc_value);
	result = setsockopt(fd, IPPROTO_SCTP, SCTP_PR_SUPPORTED, &params, len);
	if (result < 0) {
		perror("setsockopt");
		exit(-1);
	}
}
int sockopt_prsctp_support(int fd, int asoc_id)
{
	printf("[Start SCTP_PR_SUPPORTED]=======\n");
	my_get_prsctp_support(fd, asoc_id);
	my_get_prsctp_support(fd, 0);

	my_set_prsctp_support(fd, asoc_id, 0);
	my_get_prsctp_support(fd, asoc_id);
	my_get_prsctp_support(fd, 0);

	my_set_prsctp_support(fd, 0, 0);
	my_get_prsctp_support(fd, asoc_id);
	my_get_prsctp_support(fd, 0);

	my_set_prsctp_support(fd, 0, 1);
	my_get_prsctp_support(fd, asoc_id);
	my_get_prsctp_support(fd, 0);

	my_set_prsctp_support(fd, asoc_id, 1);
	my_get_prsctp_support(fd, asoc_id);
	my_get_prsctp_support(fd, 0);
	printf("[End SCTP_PR_SUPPORTED]=======\n\n");
}

int my_get_prinfo_default(int fd, int asoc_id)
{
	struct sctp_default_prinfo params;
	socklen_t len;
	int result;

	len = sizeof(params);
	memset(&params, 0, len);
	params.pr_assoc_id = asoc_id;

	result = getsockopt(fd, IPPROTO_SCTP, SCTP_DEFAULT_PRINFO, &params, &len);
	if (result < 0) {
		perror("getsockopt");
		exit(-1);
	}
	if (asoc_id)
		printf("    Get ASOC P %x, V %d\n", params.pr_policy, params.pr_value);
	else
		printf("    Get EP P %x, V %d\n", params.pr_policy, params.pr_value);
}
int my_set_prinfo_default(int fd, int asoc_id, int asoc_policy, int asoc_value)
{
	struct sctp_default_prinfo params;
	socklen_t len;
	int result;

	len = sizeof(params);
	memset(&params, 0, len);
	params.pr_assoc_id = asoc_id;
	params.pr_policy = asoc_policy;
	params.pr_value = asoc_value;
	printf("[Set Asoc ID %d, P %x, V %d  ]\n", params.pr_assoc_id, params.pr_policy, params.pr_value);
	result = setsockopt(fd, IPPROTO_SCTP, SCTP_DEFAULT_PRINFO, &params, len);
	if (result < 0) {
		perror("setsockopt");
		exit(-1);
	}
}
int sockopt_prinfo_default(int fd, int asoc_id)
{
	struct sctp_default_prinfo params;
	socklen_t len;
	int result;

	printf("[Start SCTP_PRINFO_DEFAULT]=======\n");
	my_get_prinfo_default(fd, asoc_id);

	my_set_prinfo_default(fd, asoc_id, SCTP_PR_SCTP_TTL, 10000);
	my_get_prinfo_default(fd, asoc_id);
	my_get_prinfo_default(fd, 0);

	my_set_prinfo_default(fd, 0, SCTP_PR_SCTP_TTL, 20000);
	my_get_prinfo_default(fd, asoc_id);
	my_get_prinfo_default(fd, 0);

	my_set_prinfo_default(fd, asoc_id, SCTP_PR_SCTP_RTX, 1);
	my_get_prinfo_default(fd, asoc_id);
	my_get_prinfo_default(fd, 0);

	my_set_prinfo_default(fd, 0, SCTP_PR_SCTP_RTX, 2);
	my_get_prinfo_default(fd, asoc_id);
	my_get_prinfo_default(fd, 0);

	my_set_prinfo_default(fd, asoc_id, SCTP_PR_SCTP_PRIO, 100);
	my_get_prinfo_default(fd, asoc_id);
	my_get_prinfo_default(fd, 0);

	my_set_prinfo_default(fd, 0, SCTP_PR_SCTP_PRIO, 200);
	my_get_prinfo_default(fd, asoc_id);
	my_get_prinfo_default(fd, 0);

	my_set_prinfo_default(fd, asoc_id, SCTP_PR_SCTP_NONE, 2000);
	my_get_prinfo_default(fd, asoc_id);
	my_get_prinfo_default(fd, 0);
	printf("[End SCTP_PRINFO_DEFAULT]=======\n\n");
}


int my_pr_assocstatus(int fd, int asoc_id, int asoc_policy)
{
	struct sctp_prstatus params;
	socklen_t len;
	int result;

	len = sizeof(params);
	memset(&params, 0, len);
	params.sprstat_assoc_id = asoc_id;
	params.sprstat_policy = asoc_policy;
	result = getsockopt(fd, IPPROTO_SCTP, SCTP_PR_ASSOC_STATUS, &params, &len);
	if (result < 0) {
		perror("getsockopt");
		exit(-1);
	}
	printf("   Get policy %d, unsent %d, sent %d\n", params.sprstat_policy, params.sprstat_abandoned_unsent, params.sprstat_abandoned_sent);
}
int sockopt_pr_assocstatus(int fd, int asoc_id)
{
	printf("[Start SCTP_ASSOCSTATUS]=======\n");
	my_pr_assocstatus(fd, asoc_id, SCTP_PR_SCTP_TTL);
	my_pr_assocstatus(fd, asoc_id, SCTP_PR_SCTP_RTX);
	my_pr_assocstatus(fd, asoc_id, SCTP_PR_SCTP_PRIO);
	my_pr_assocstatus(fd, asoc_id, SCTP_PR_SCTP_NONE);
	printf("[End SCTP_PR_ASSOCSTATUS]=======\n\n");
}

int ttl_setup(struct sctp_sndrcvinfo *sinfo, int ttl)
{
	sinfo->sinfo_flags = SCTP_PR_SCTP_TTL;
	sinfo->sinfo_timetolive = ttl;
	printf("setup the ttl value is %d \n", ttl);

	return 0;
}

int rtx_setup(struct sctp_sndrcvinfo *sinfo, int count)
{
	sinfo->sinfo_flags = SCTP_PR_SCTP_RTX;
	sinfo->sinfo_timetolive = count;
	printf("setup the count is %d \n", count);

	return 0;
}

int prio_setup(struct sctp_sndrcvinfo *sinfo, int priority, int fd, int n)
{
	if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &n, sizeof(n)) == -1) {
		perror("setsockopt");
		exit(-1);
	}
	sinfo->sinfo_flags = SCTP_PR_SCTP_PRIO;
	sinfo->sinfo_timetolive = priority;
	printf("setup the priority is %d \n", priority);

	return 0;
}
int firewall_config()
{
	system("iptables -A OUTPUT -p sctp --chunk-types only DATA -o veth0 -j DROP");
	//system("ip netns exec sctp_s iptables -A INPUT -p sctp --chunk-types only DATA -i veth0 -j DROP");
	return 0;
}
int firewall_clear()
{
	system("iptables -F");
	//system("ip netns exec sctp_s iptables -F");
	return 0;
}

static int do_perform(int policy)
{
	int fd, msg_flags, send_len, recv_len, total_len = 0;
	time_t start, end;
	char addr_str[30];
	char buf[BUFFERSIZE];

	struct sctp_sndrcvinfo sri;
	struct addrinfo *res = NULL;
	struct sockaddr_storage peeraddr;
	socklen_t len = sizeof(peeraddr);

	struct sctp_event_subscribe evnts;
	struct sockaddr_storage addr_cli, addr_ser;

	/*sctp connect*/
	memset(&addr_ser, 0, sizeof(addr_ser));
	sprintf(addr_str, "172.16.%d.%d", 3, 1);
	getaddrinfo(addr_str, "8888", NULL, &res);
	memcpy(&addr_ser, res->ai_addr, res->ai_addrlen);
	((struct sockaddr_in *)(&addr_ser))->sin_port = htons(8888);

	memset(&addr_cli, 0, sizeof(addr_cli));
	sprintf(addr_str, "172.16.%d.%d", 3, 2);
	getaddrinfo(addr_str, "8888", NULL, &res);
	memcpy(&addr_cli, res->ai_addr, res->ai_addrlen);
	((struct sockaddr_in *)(&addr_cli))->sin_port = htons(8888);

	if ((fd = socket(res->ai_family, SOCK_SEQPACKET, IPPROTO_SCTP)) == -1){
		perror("socket");
		exit(-1);
	}
	if (bind(fd, (struct sockaddr *)&addr_cli, sizeof(addr_cli)) == -1){
		perror("bind");
		exit(-1);
	}
	memset(&evnts, 0, sizeof(evnts));
	evnts.sctp_data_io_event = 1;
	if (setsockopt(fd, IPPROTO_SCTP, SCTP_EVENTS, &evnts,sizeof(evnts)) == -1){
		perror("setsockopt");
		exit(-1);
	}
	if (connect(fd, (struct sockaddr *)&addr_ser, sizeof(addr_ser)) == -1){
		perror("connect");
		exit(-1);
	}

	/*create assocs*/
	create_assocs();
	/*get asoc id*/
	int asoc_id = get_asoc_id(fd);
	printf("::::::::::::::::::::::::: TEST POLICY %d :::::::::::::::::::::::\n\n", policy);

	if (policy == -1) {
		sockopt_prsctp_support(fd, asoc_id);

		sockopt_prinfo_default(fd, asoc_id);

		sockopt_pr_assocstatus(fd, asoc_id);
	}

	start = time(NULL);
	int i = 0, loops=1;
	if (policy == 3)
		loops=3;
	while(i<loops){
		struct sctp_sndrcvinfo sinfo;

		memset(&sinfo, 0, sizeof(sinfo));
		if(policy == 0) {
			printf("[Start SCTP_MSG_EXPIRE]=======\n");
			/*
			my_set_prsctp_support(fd, asoc_id, 0);
			*/
			my_set_prsctp_support(fd, asoc_id, 1);
			sinfo.sinfo_timetolive = 5000;
			firewall_config();
		} else if(policy == 1) {
			printf("[Start SCTP_PR_TTL]=======\n");
			my_set_prsctp_support(fd, asoc_id, 1);
			ttl_setup(&sinfo, 15000);
			firewall_config();
		} else if(policy == 2) {
			printf("[Start SCTP_PR_RTX]=======\n");
			my_set_prsctp_support(fd, asoc_id, 1);
			rtx_setup(&sinfo, 2);
			/*
			rtx_setup(&sinfo, 5);
			*/
			firewall_config();
		} else if (policy == 3){
			printf("[Start SCTP_PR_PRIO]=======\n");
			my_set_prsctp_support(fd, asoc_id, 1);
			prio_setup(&sinfo, 5-i, fd, 1500);
			firewall_config();
		}

		if ((send_len = sctp_sendmsg(fd, buf, BUFFERSIZE, 
			(struct sockaddr *)&addr_ser, sizeof(addr_ser), 0, sinfo.sinfo_flags, 0, sinfo.sinfo_timetolive, 0)) == -1) {
			perror("sctp_sendmsg");
			exit(-1);
		}
		total_len += send_len;
		i++;
	}
	end = time(NULL);
	printf("time is %lf, send pkt is %d\n", difftime(end, start), total_len);
	if (policy != -1)
		printf("after see the FOR CUM TSN packet in sctp_s\n");
	printf("press Enter to continue\n");
	getchar();
	firewall_clear();
	sockopt_pr_assocstatus(fd, asoc_id);
	close(fd);
	sleep(5);
	return 0;
}
int main()
{
	printf("\n **** RFC 7496 TEST ****\n\n");
	/*
	int policy = 0;
	int policy = 1;
	int policy = 2;
	int policy = 3;
	*/

	do_perform(-1);
	do_perform(0);
	do_perform(1);
	do_perform(2);
	do_perform(3);

	printf("\n **** END ****\n\n");
}

